/**
* Copyright (c) 2009, Marc 'Foddex' Oude Kotte <foddex@foddex.net>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Marc 'Foddex' Oude Kotte nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Marc 'Foddex' Oude Kotte ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Marc 'Foddex' Oude Kotte BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef FODDEX_JSON_RPC_TRANSPORT_CURL_H
#define FODDEX_JSON_RPC_TRANSPORT_CURL_H

// make sure curl/curl.h is on your include path!
#include <curl/curl.h>
#include "json-rpc.h"

namespace json {

	/** Transport based on CURL. */
	class curl_transport : public transport {
		private:
			CURL*			mCurlHandle;
			string			mUrl;
			static size_t data_function( void *ptr, size_t size, size_t nmemb, void *stream) {
				string* data = static_cast<string*>( stream );
				size_t blockSize = size*nmemb;
				data->append( static_cast<const char*>( ptr ), blockSize );
				return blockSize;
			}
			static size_t post_function( void *ptr, size_t size, size_t nmemb, void *stream) {
				string* call_string = static_cast<string*>( stream );
				size_t blockSize = min<size_t>( size*nmemb, call_string->size() );
				memcpy( ptr, &(*call_string)[0], blockSize );
				call_string->erase( 0, blockSize );
				return blockSize;
			}
		protected:
			string execute( const string& call_string ) {
				string result;
				// make sure the data ends up with us
				curl_easy_setopt( mCurlHandle, CURLOPT_WRITEFUNCTION, &curl_transport::data_function );
				curl_easy_setopt( mCurlHandle, CURLOPT_WRITEDATA, &result );
				// setup post
				curl_easy_setopt( mCurlHandle, CURLOPT_POSTFIELDS, NULL );
				curl_easy_setopt( mCurlHandle, CURLOPT_POSTFIELDSIZE, call_string.size() );
				curl_easy_setopt( mCurlHandle, CURLOPT_READFUNCTION, &curl_transport::post_function );
				curl_easy_setopt( mCurlHandle, CURLOPT_READDATA, &call_string );
				curl_easy_setopt( mCurlHandle, CURLOPT_SSL_VERIFYPEER, 0 );
				curl_easy_setopt( mCurlHandle, CURLOPT_SSL_VERIFYHOST, 0 );
				// perform request!
				CURLcode res = curl_easy_perform( mCurlHandle );
				if (res!=CURLE_OK)
					throw json_rmi_exception( "failed to perform the post using curl" );
				return result;
			}
		public:
			/** Constructor, sets up new CURL transport for Json web service located on rpc_url */
			curl_transport( const string& rpc_url ) 
			:	mCurlHandle( curl_easy_init() )
			,	mUrl( rpc_url ) {
				if (!mCurlHandle)
					throw json_rmi_exception( "failed to obtain curl handle" );
				curl_easy_setopt( mCurlHandle, CURLOPT_HEADER, 0 );					// don't include header in output
				curl_easy_setopt( mCurlHandle, CURLOPT_FOLLOWLOCATION, 1 );			// automatically follow HTTP 300 responses
				curl_easy_setopt( mCurlHandle, CURLOPT_MAXREDIRS, 8 );				// never follow a HTTP 300 more than 8 times
				curl_easy_setopt( mCurlHandle, CURLOPT_CONNECTTIMEOUT, 10 );		// 10 seconds connection timeout
				curl_easy_setopt( mCurlHandle, CURLOPT_FAILONERROR, 1 );
				curl_easy_setopt( mCurlHandle, CURLOPT_URL, mUrl.c_str() );			// setup url we'll be calling
				curl_easy_setopt( mCurlHandle, CURLOPT_POST, 1 );					// tell curl we want posts
			}
			/** Destructor, cleans up curl handle */
			~curl_transport() {
				if (mCurlHandle)
					curl_easy_cleanup( mCurlHandle );
			}
			/** Returns CURL handle in case you want to setup your own settings before performing operations. 
			 *	Do note that changing the CURLOPT_WRITEFUNCTION and/or CURLOPT_WRITEDATA settings has no effect,
			 *	as they are set in the internal execute() function everytime the transport object is used!
			 */
			CURL* get_curl_handle() { 
				return mCurlHandle; 
			}
	};

} // json

#endif
