/**
* Copyright (c) 2009, Marc 'Foddex' Oude Kotte <foddex@foddex.net>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of Marc 'Foddex' Oude Kotte nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY Marc 'Foddex' Oude Kotte ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Marc 'Foddex' Oude Kotte BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef FODDEX_JSON_RPC_H
#define FODDEX_JSON_RPC_H

#include "json-templates.h"

namespace json {

	/** 
	 * This code is based on working draft 7:
	 *		http://json-rpc.org/wd/JSON-RPC-1-1-WD-20060807.html
	 */

	/** Runtime rmi exception */
	class json_rmi_exception : public json_exception {
		public:
			json_rmi_exception( const string& message ) : json_exception( message ) {}
	};

	/** Expected Json rpc version. */
	static const char* json_rpc_version = "1.1";

	/** Interface for handling transport of the rpc call and result. */
	class transport {
		private:
			__int64 mNextId;
		protected:
			virtual string execute( const string& call_string ) = 0;
		public:
			transport() : mNextId( 0 ) {}
			json_integer* get_id() { return new json_integer( ++mNextId ); }
			string execute( json_object* call ) {
				return execute( call->written() );
			}
	};

	/** Determines if type requires any serialization. */
	template<typename _T>	struct type_requires_serialization			{ enum { value = true }; };
	template<>				struct type_requires_serialization<void>	{ enum { value = false }; };

	/** Result type serialization */
	template<typename _T>
	struct result_type_serialization {
		_T operator()( const json_block* unparsed_value ) {
			_T value;
			serialize<_T>::read( value, unparsed_value );
			return value;
		}
	};
	template<>
	struct result_type_serialization<void> {
		void operator()( const json_block* ) {
		}
	};

	/** Object that can perform json based rpc calls. */
	class rpc {

		/** Auto-free object */
		struct auto_free {
			json_block* Block;
			auto_free( json_block* block ) : Block( block ) {}
			~auto_free() { if (Block) delete Block; }
		};

		private:
			transport*			mTransport;

		protected:
			json_object* create_call( const string& method_name, json_object* params, bool have_id ) {
				json_object* obj = new json_object;
				obj->set( "version", new json_string( json_rpc_version ) );
				obj->set( "method", new json_string( method_name ) );
				if (params)
					obj->set( "params", params );
				if (have_id)
					obj->set( "id", get_transport()->get_id() );
				return obj;
			}

			template<typename _R>
			_R do_call( const string& method_name, json_object* params ) {
				auto_free af( params );			// will auto free on exception and end-of-function
				json_block* returnBlock = json_parser( get_transport()->execute( create_call( method_name, params, type_requires_serialization<_R>::value ) ) ).get();
				auto_free af2( returnBlock );	// will auto free on exception and end-of-function
				json_object* returnObject = dynamic_cast<json_object*>( returnBlock );
				if (!returnObject)
					throw json_rmi_exception( "expected a json object as result" );
				json_block* errorObject = returnObject->get_value( "error" );
				if (!dynamic_cast<json_null*>( errorObject )) {
					json_string* errorString = dynamic_cast<json_string*>( errorObject );
					if (errorString)
						throw json_rmi_exception( "json error: " + errorString->get_value() );
					throw json_rmi_exception( "json error: <unknown reason>" );
				}
				// XXX check id ?
				return result_type_serialization<_R>()( returnObject->get_value( "result" ) );
			}

		public:
			/** Constructor, takes means of transport (e.g. tcp connector) for handling RPC. */
			rpc( transport* transport ) : mTransport( transport ) {
				if (!mTransport)
					throw json_exception( "no transport specified to rpc constructor" );
			}

			/** Returns transport being used. */
			transport* get_transport() { return mTransport; }
			const transport* get_transport() const { return mTransport; }
	
		public:
			// 0 parameters
			template<typename _R>
			_R call( const string& method_name, const char*[]=NULL ) {
				return do_call<_R>( method_name, NULL );
			}
			// 1 parameters
			template<typename _R, typename _P0>
			_R call( const string& method_name, const char* params[], const _P0& p0 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 2 parameters
			template<typename _R, typename _P0, typename _P1>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 3 parameters
			template<typename _R, typename _P0, typename _P1, typename _P2>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1, const _P2& p2 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				paramsObject->set( params ? params[2] : "p2", serialize<_P2>::write( p2 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 4 parameters
			template<typename _R, typename _P0, typename _P1, typename _P2,  typename _P3>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1, const _P2& p2, const _P3& p3 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				paramsObject->set( params ? params[2] : "p2", serialize<_P2>::write( p2 ) );
				paramsObject->set( params ? params[3] : "p3", serialize<_P3>::write( p3 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 5 parameters
			template<typename _R, typename _P0, typename _P1, typename _P2,  typename _P3, typename _P4>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1, const _P2& p2, const _P3& p3, const _P4& p4 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				paramsObject->set( params ? params[2] : "p2", serialize<_P2>::write( p2 ) );
				paramsObject->set( params ? params[3] : "p3", serialize<_P3>::write( p3 ) );
				paramsObject->set( params ? params[4] : "p4", serialize<_P4>::write( p4 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 6 parameters
			template<typename _R, typename _P0, typename _P1, typename _P2,  typename _P3, typename _P4, typename _P5>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1, const _P2& p2, const _P3& p3, const _P4& p4, const _P5& p5 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				paramsObject->set( params ? params[2] : "p2", serialize<_P2>::write( p2 ) );
				paramsObject->set( params ? params[3] : "p3", serialize<_P3>::write( p3 ) );
				paramsObject->set( params ? params[4] : "p4", serialize<_P4>::write( p4 ) );
				paramsObject->set( params ? params[5] : "p5", serialize<_P5>::write( p5 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 7 parameters
			template<typename _R, typename _P0, typename _P1, typename _P2,  typename _P3, typename _P4, typename _P5, typename _P6>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1, const _P2& p2, const _P3& p3, const _P4& p4, const _P5& p5, const _P6& p6 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				paramsObject->set( params ? params[2] : "p2", serialize<_P2>::write( p2 ) );
				paramsObject->set( params ? params[3] : "p3", serialize<_P3>::write( p3 ) );
				paramsObject->set( params ? params[4] : "p4", serialize<_P4>::write( p4 ) );
				paramsObject->set( params ? params[5] : "p5", serialize<_P5>::write( p5 ) );
				paramsObject->set( params ? params[6] : "p6", serialize<_P6>::write( p6 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 8 parameters
			template<typename _R, typename _P0, typename _P1, typename _P2,  typename _P3, typename _P4, typename _P5, typename _P6, typename _P7>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1, const _P2& p2, const _P3& p3, const _P4& p4, const _P5& p5, const _P6& p6, const _P7& p7 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				paramsObject->set( params ? params[2] : "p2", serialize<_P2>::write( p2 ) );
				paramsObject->set( params ? params[3] : "p3", serialize<_P3>::write( p3 ) );
				paramsObject->set( params ? params[4] : "p4", serialize<_P4>::write( p4 ) );
				paramsObject->set( params ? params[5] : "p5", serialize<_P5>::write( p5 ) );
				paramsObject->set( params ? params[6] : "p6", serialize<_P6>::write( p6 ) );
				paramsObject->set( params ? params[7] : "p7", serialize<_P7>::write( p7 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 9 parameters
			template<typename _R, typename _P0, typename _P1, typename _P2,  typename _P3, typename _P4, typename _P5, typename _P6, typename _P7, typename _P8>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1, const _P2& p2, const _P3& p3, const _P4& p4, const _P5& p5, const _P6& p6, const _P7& p7, const _P8& p8 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				paramsObject->set( params ? params[2] : "p2", serialize<_P2>::write( p2 ) );
				paramsObject->set( params ? params[3] : "p3", serialize<_P3>::write( p3 ) );
				paramsObject->set( params ? params[4] : "p4", serialize<_P4>::write( p4 ) );
				paramsObject->set( params ? params[5] : "p5", serialize<_P5>::write( p5 ) );
				paramsObject->set( params ? params[6] : "p6", serialize<_P6>::write( p6 ) );
				paramsObject->set( params ? params[7] : "p7", serialize<_P7>::write( p7 ) );
				paramsObject->set( params ? params[8] : "p8", serialize<_P8>::write( p8 ) );
				return do_call<_R>( method_name, paramsObject );
			}
			// 10 parameters
			template<typename _R, typename _P0, typename _P1, typename _P2,  typename _P3, typename _P4, typename _P5, typename _P6, typename _P7, typename _P8, typename _P9>
			_R call( const string& method_name, const char* params[], const _P0& p0, const _P1& p1, const _P2& p2, const _P3& p3, const _P4& p4, const _P5& p5, const _P6& p6, const _P7& p7, const _P8& p8, const _P9& p9 ) {
				json_object* paramsObject = new json_object();
				paramsObject->set( params ? params[0] : "p0", serialize<_P0>::write( p0 ) );
				paramsObject->set( params ? params[1] : "p1", serialize<_P1>::write( p1 ) );
				paramsObject->set( params ? params[2] : "p2", serialize<_P2>::write( p2 ) );
				paramsObject->set( params ? params[3] : "p3", serialize<_P3>::write( p3 ) );
				paramsObject->set( params ? params[4] : "p4", serialize<_P4>::write( p4 ) );
				paramsObject->set( params ? params[5] : "p5", serialize<_P5>::write( p5 ) );
				paramsObject->set( params ? params[6] : "p6", serialize<_P6>::write( p6 ) );
				paramsObject->set( params ? params[7] : "p7", serialize<_P7>::write( p7 ) );
				paramsObject->set( params ? params[8] : "p8", serialize<_P8>::write( p8 ) );
				paramsObject->set( params ? params[9] : "p9", serialize<_P9>::write( p9 ) );
				return do_call<_R>( method_name, paramsObject );
			}
	};

} // json

#endif
